/*
*  File             : pipe.c 
*  Version          : 1.0
*  Written by       : alinden@netcologne.de
*  Date             : Jul 15 1998
*  Compiler         : gcc 2.7.2 on Linux 2.0.33, libc5
*  Portability      : Should be POSIX compliant
*  Copyright:       : GPL
*
*  DESCRIPTION: Exchange information with the crontab command
*/

#include "conf.h"
#include "crond.h"
#include "extern.h"

void syslogw(char *, int);
void update_ctab(char *);
char *cpstr(size_t, char *, ...);
void restore_sigs(void);
void msgout(char *);
void run_proc(CONFIG *);
int cmd_queued(CONFIG *); 

void sig_alrm(int signo) { return; }
int sig_pipe(int signo) { exit(0); }

int open_pipe(char *ctab, char *username)  /* open output pipe to user */
{
	int wfifo;
	
	signal(SIGALRM, (void *)sig_alrm);
	alarm(6);

	wfifo = open(cpstr(256, "./%s", ctab), O_WRONLY);

	alarm(0);

	if(wfifo == -1)
	{
		if(errno == EINTR)
			syslogw(cpstr(200, "pipe to user %s timed out", username), LOG_WARNING);
		else
			syslogw(cpstr(200, "cannot open output pipe to user %s:%%m", username), LOG_WARNING);
	}

	return(wfifo);
}

void outfifo(char *ctab, char *cbuf)   /* send joblist to pipe, ctab = FIFO name, cbuf = regexp */
{
	pid_t pid;
	char *username;
	CONFIG *tbuf;
	FILE *ffifo=NULL;
	char cmain;
	char *tmstr = "%b %d %a %H:%M";
	time_t ltm,lth=0;
	char tmbuf[20];
	int wfifo,jobs,fjobs,lstall,i;
	
	if((pid = fork()) < 0)
	{
		syslogw("can't fork", LOG_WARNING);
		return;
	}

	if(pid)
		return;

	restore_sigs();

	lstall = 0;

	if(strcmp(ctab+8, NCTROOT) == 0)
	{
		username = "root";
		cmain = 1;
	}
	else
	{
		username = ctab+8;
		cmain = 0;
	}	

	if(cmain && strcmp(cbuf, "*all*") == 0)
		lstall = 1;

	if((wfifo = open_pipe(ctab, username)) == -1 || (ffifo = fdopen(wfifo, "w")) == NULL)
		exit(0);

	tbuf = cblock;

	if(lstall == 0)
	{
		while(tbuf != NULL)  /* find user */
		{
			if((cmain && tbuf->cmain) || 
			  (! cmain && ! tbuf->cmain && strcmp(username, tbuf->username) == 0))
				break;

			tbuf = tbuf->nuser;
		}
	}
	
	signal(SIGPIPE, (void *)sig_pipe);

	if(tbuf == NULL)
	{
		if(lstall == 0) 
			fprintf(ffifo, "No jobs for user %s\n", username);
		else
			fprintf(ffifo, "No jobs\n");

		fclose(ffifo);
		exit(0);
	}

	if(lstall == 0)
		fprintf(ffifo, "Jobs for user %s\n", username);

	time(&ltm);
		
	jobs=fjobs=0;
	while(1)
	{
		jobs++;

		if(lstall == 0)
		{
			if(! (tbuf->jobid != NULL && fnmatch(cbuf, tbuf->jobid, 0) == 0) && 
			fnmatch(cbuf, tbuf->cmdstr, 0))
			{
				tbuf = tbuf->nblock;

				if(tbuf == NULL || tbuf->stblock)
					break;

				continue;	
			}
		}
		
		if(lstall == 0 || fjobs)
			fprintf(ffifo, "-----------------------------------------\n");

		fjobs = 1;	
		
		fprintf(ffifo, "Number : %d\n", jobs);
		fprintf(ffifo, "Command: %s\n", tbuf->cmdstr);
		
		if(cmain)
			fprintf(ffifo, "User   : %s\n", tbuf->username); 
		
		fprintf(ffifo, "Subject: %s\n", (tbuf->jobid)?(tbuf->jobid):("<default>"));
		
		fprintf(ffifo, "Options: ");

		i = 0;

		if(tbuf->period)
		{
			fputc('U', ffifo);
			i = 1;
		}	
		if(tbuf->dopt)
		{
			fputc('D', ffifo);
			i = 1;
		}

		if(tbuf->bopt)
		{
			fputc('B', ffifo);	
			i = 1;
		}

		if(i == 0)
			fprintf(ffifo, "<none>");

		fputc('\n', ffifo);
		
		fprintf(ffifo, "Status : ");

		if(tbuf->pid)
		{
			fprintf(ffifo, "Running\n");

		} else if(cmd_queued(tbuf))
		{
			fprintf(ffifo, "Queued\n");
			
		} else
			fprintf(ffifo, "Waiting\n");

		fprintf(ffifo, "Last   : ");

		if(tbuf->lastexec == 0)
			fprintf(ffifo, "<none>\n");
		else
		{
			strftime(tmbuf, 20, tmstr, localtime(&tbuf->lastexec));
			lth = (ltm-tbuf->lastexec)/(60*60);
			fprintf(ffifo, "%s  +(%dh %.0fm)\n", tmbuf, (int)lth, (((float)ltm-(float)tbuf->lastexec)/60)-(float)lth*60);
		}

		fprintf(ffifo, "Next   : ");

		if(tbuf->nextexec == -1)
			fprintf(ffifo, "<?>\n");
		else
		{
			strftime(tmbuf, 20, tmstr, localtime(&tbuf->nextexec));
			lth = (tbuf->nextexec-ltm)/(60*60);
			fprintf(ffifo, "%s  -(%dh %.0fm)\n", tmbuf, (int)lth, (((float)tbuf->nextexec-(float)ltm)/60)-(float)lth*60);
		}
		
		tbuf = tbuf->nblock;

		if(tbuf == NULL || (! lstall && tbuf->stblock))
			break;

	}

	if(lstall == 0 && fjobs == 0)
		fprintf(ffifo, "No match with %s\n", cbuf);
		
	fclose(ffifo);

	exit(0);
}

void sndmsg(char *message, char *ctab, char *username)   /* send a message to the pipe */
{
	pid_t pid;
	int wfifo;
	
	if((pid = fork()) < 0)
	{
		syslogw("can't fork", LOG_WARNING);
		return;
	}

	if(pid)
		return;

	restore_sigs();
	
	if((wfifo = open_pipe(ctab, username)) == -1)
		exit(0);
		
	signal(SIGPIPE, (void *)sig_pipe);
	write(wfifo, message, strlen(message));
	close(wfifo);
	
	exit(0);
}

void runcmd(char *ctab, char *cbuf)   /* run command, ctab = FIFO name, cbuf = command */
{
	char *username;
	CONFIG *tbuf;
	char cmain;
	int job,i;
	
	if(strcmp(ctab+8, NCTROOT) == 0)
	{
		username = "root";
		cmain = 1;
	}
	else
	{
		username = ctab+8;
		cmain = 0;
	}	

	tbuf = cblock;

	while(tbuf != NULL)  /* find user */
	{
		if((cmain && tbuf->cmain) || 
		  (! cmain && ! tbuf->cmain && strcmp(username, tbuf->username) == 0))
			break;

		tbuf = tbuf->nuser;
	}

	if(tbuf == NULL)
	{
		sndmsg(cpstr(256, "No jobs for user %s\n", username), ctab, username);
		return;
	}

	for(i = 0; cbuf[i] >= '0' && cbuf[i] <= '9'; i++);

	if(cbuf[i])  /* alphanumeric input */ 
	{
		do
		{
			if(tbuf->jobid != NULL && strcmp(tbuf->jobid, cbuf) == 0)
				break;

			tbuf = tbuf->nblock;

			if(tbuf != NULL && tbuf->stblock)
				tbuf = NULL;
				
		} while(tbuf != NULL);
	}
	else
	{
		job = atoi(cbuf);

		if(job)
		{
			for(i = 1; i < job; i++)
			{
				tbuf = tbuf->nblock;

				if(tbuf != NULL && tbuf->stblock)
					tbuf = NULL;

				if(tbuf == NULL)
					break;
			}
		}
		else 
			tbuf = NULL;
	}

	if(tbuf == NULL)
	{
		sndmsg(cpstr(256, "No such job (%s)\n", cbuf), ctab, username);
		return;
	}

	if(tbuf->jobid)
		cbuf = tbuf->jobid;

	if(tbuf->pid)
	{
		sndmsg(cpstr(256, "Job already running (%s)\n", cbuf), ctab, username);
		return;
	}

	if(cmd_queued(tbuf))
	{
		sndmsg(cpstr(256, "Job is queued (%s)\n", cbuf), ctab, username);
		return;
	}

	run_proc(tbuf);

	sndmsg(cpstr(256, "Command %s started\n", cbuf), ctab, username);
}

void read_fifo(void)    /* read a command from pipe */
{
	ssize_t nrd;
	char *cbuf,*cbuf2=NULL;
	char ctab[201];

	if((nrd = read(rwfifo, ctab, 201)) < 0)
	{
		syslogw("while reading from pipe: %m", LOG_WARNING);
		return;
	}
	
	if(nrd > 200)
	{
		syslogw("input to pipe too long (max. 200 chars)", LOG_WARNING);
		return;
	}

	*(ctab+nrd) = 0;
	*(ctab+strcspn(ctab, "\n")) = 0;

	if(strncmp(ctab, "@output-", 8) == 0)    /* reply requested */
	{
		/* find command */
		
		for(cbuf = ctab+8; (cbuf = strstr(cbuf+1, " @")) != NULL; cbuf2 = cbuf);

		if(cbuf2 != NULL)
		{
			cbuf = cbuf2;
			while(cbuf != ctab &&  ! isalnum(*(--cbuf)));

			*(cbuf+1) = 0;
			cbuf = cbuf2+2;

			if(*cbuf == 'R')  /* run command */
			{
				runcmd(ctab, cbuf+1);
				return;
			}

			outfifo(ctab, cbuf);
			return;
		} 
		else
		{
			cbuf = "*";
			outfifo(ctab, cbuf);
			return;
		}
	}

	update_ctab(ctab);
}
